/*
 * Jalview - A Sequence Alignment Editor and Viewer
 * Copyright (C) 2006 AM Waterhouse, J Procter, G Barton, M Clamp, S Searle
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package jalview.commands;

import jalview.datamodel.*;

/**
 *
 * <p>Title: EditCommmand</p>
 *
 * <p>Description: Essential information for performing
 * undo and redo for cut/paste insert/delete gap
 * which can be stored in the HistoryList </p>
 *
 * <p>Copyright: Copyright (c) 2006</p>
 *
 * <p>Company: Dundee University</p>
 *
 * @author not attributable
 * @version 1.0
 */
public class EditCommand implements CommandI
{
  public static String INSERT_GAP = "InsertGap";
  public static String DELETE_GAP = "DeleteGap";
  public static String CUT = "Cut";
  public static String PASTE = "Paste";

  Edit[] edits;

  String description;

  public EditCommand()
  {}

  public EditCommand(String description)
  {
    this.description = description;
  }

  public EditCommand(String description,
                     String command,
                     SequenceI[] seqs,
                     int position,
                     int number,
                     char gapChar)
  {
    this.description = description;

    if (command.equalsIgnoreCase(INSERT_GAP)
        || command.equalsIgnoreCase(DELETE_GAP))
    {
      edits = new Edit[] { new Edit(command, seqs, position, number, gapChar)};
    }

    performEdit(0);
  }

  public EditCommand( String description,
                      String command,
                      SequenceI[] seqs,
                      int position,
                      int number,
                      AlignmentI al)
   {
     this.description = description;
     if ( command.equalsIgnoreCase(CUT) || command.equalsIgnoreCase(PASTE))
     {
       edits = new Edit[]{new Edit(command, seqs, position, number, al)};
     }

     performEdit(0);
  }


  public String getDescription()
  {
    return description;
  }

  public int getSize()
  {
    return edits==null?0:edits.length;
  }

  public AlignmentI getAlignment()
  {
    return edits[0].al;
  }


  public void appendEdit(String command,
                         SequenceI[] seqs,
                         int position,
                         int number,
                         char gapChar,
                         boolean performEdit)
  {
    Edit edit = new Edit(command, seqs, position, number, gapChar);

    if (edits != null)
    {
      Edit[] temp = new Edit[edits.length + 1];
      System.arraycopy(edits, 0, temp, 0, edits.length);
      edits = temp;
      edits[edits.length - 1] = edit;
    }
    else
      edits = new Edit[] {  edit  };

    if (performEdit)
      performEdit(edits.length - 1);
  }

  void performEdit(int commandIndex)
  {
    int eSize = edits.length;
    for (int e = commandIndex; e < eSize; e++)
    {
      if (edits[e].command.equals(INSERT_GAP))
      {
        insertGap(edits[e]);
      }
      else if (edits[e].command.equals(DELETE_GAP))
      {
        deleteGap(edits[e]);
      }
      else if(edits[e].command.equals(CUT))
      {
        cut(edits[e]);
      }
      else if(edits[e].command.equals(PASTE))
      {
        paste(edits[e]);
      }
    }
  }

  public void doCommand()
  {
    performEdit(0);
  }

  public void undoCommand()
  {
    int e = 0, eSize = edits.length;
    for (e = eSize-1; e > -1; e--)
    {
      if (edits[e].command.equals(INSERT_GAP))
      {
        deleteGap(edits[e]);
      }
      else if (edits[e].command.equals(DELETE_GAP))
      {
        insertGap(edits[e]);
      }
      else if (edits[e].command.equals(CUT))
      {
        paste(edits[e]);
      }
      else if (edits[e].command.equals(PASTE))
      {
        cut(edits[e]);
      }
    }
  }

  void insertGap(Edit command)
  {
    for(int s=0; s<command.seqs.length; s++)
    {
        command.seqs[s].insertCharAt(command.position,
                                     command.number,
                                     command.gapChar);
    }
  }

  void deleteGap(Edit command)
  {
    for (int s = 0; s < command.seqs.length; s++)
    {
      command.seqs[s].deleteChars(command.position, command.position+command.number);
    }
  }

  void cut(Edit command)
  {
    command.string = new String [command.seqs.length];

    for(int i=0; i<command.seqs.length; i++)
    {
      if(command.seqs[i].getLength()>command.position)
      {
        command.string[i] = command.seqs[i].getSequence(command.position,
            command.position + command.number);

        command.seqs[i].deleteChars(command.position,
                                    command.position + command.number);
      }

      if(command.seqs[i].getLength()<1)
      {
        command.al.deleteSequence(command.seqs[i]);
      }
    }
  }

  void paste(Edit command)
  {
    StringBuffer tmp;
    for(int i=0; i<command.seqs.length; i++)
    {
      if(command.seqs[i].getLength()<1)
      {
        // ie this sequence was deleted, we need to
        // read it to the alignment
        if (command.alIndex[i] < command.al.getHeight())
          command.al.getSequences().insertElementAt(command.seqs[i],
              command.alIndex[i]);
        else
          command.al.addSequence(command.seqs[i]);
      }
      tmp = new StringBuffer(command.seqs[i].getSequence());

      if(command.string!=null && command.string[i]!=null)
      {
        if(command.position>=tmp.length())
        {
          //This occurs if padding is on, and residues
          //are removed from end of alignment
          int length = command.position-tmp.length();
          while (length > 0)
          {
            tmp.append(command.gapChar);
            length--;
          }
        }
        tmp.insert(command.position, command.string[i]);
        command.string[i] = null;
      }
      command.seqs[i].setSequence(tmp.toString());
    }

    command.string = null;
  }

  class Edit
  {
    AlignmentI al;
    String command;
    String [] string;
    SequenceI[] seqs;
    int [] alIndex;
    int position, number;
    char gapChar;

    Edit(String command,
         SequenceI[] seqs,
         int position,
         int number,
         char gapChar)
    {
      this.command = command;
      this.seqs = seqs;
      this.position = position;
      this.number = number;
      this.gapChar = gapChar;
    }


    Edit(String command,
         SequenceI[] seqs,
         int position,
         int number,
         AlignmentI al)
    {
      this.gapChar = al.getGapCharacter();
      this.command = command;
      this.seqs = seqs;
      this.position = position;
      this.number = number;
      this.al = al;
      alIndex = new int[seqs.length];
      for(int i=0; i<seqs.length; i++)
        alIndex[i] = al.findIndex(seqs[i]);
    }

  }

}
